@@ -21,25 +21,26 @@
  */
 
 #include "config.h"
-#include <stdlib.h>
-#include <string.h>
-#include <strings.h>
-#include <popt.h>
-#include <gdk/gdk.h>
-#include <libbonobo.h>
-#include <login-helper/login-helper.h>
-#include <gdk/gdkx.h>
-#include <gtk/gtk.h>
-#include <X11/Xatom.h>
-#ifdef HAVE_XFIXES
-#include <X11/extensions/Xfixes.h>
-#endif
 #include "magnifier.h"
 #include "magnifier-private.h"
 #include "zoom-region.h"
 #include "zoom-region-private.h"
+#include "gmag-cursor.h"
+#include "gmag-graphical-server.h"
 #include "GNOME_Magnifier.h"
 
+#include <stdlib.h>
+#include <string.h>
+#include <strings.h>
+
+#include <libbonobo.h>
+#include <login-helper/login-helper.h>
+
+#include <gdk-pixbuf/gdk-pixbuf-io.h>
+#include <gdk/gdkx.h>
+#include <gdk/gdk.h>
+#include <gtk/gtk.h>
+
 /* if you #define this, don't forget to set MAG_CLIENT_DEBUG env variable */
 #define DEBUG_CLIENT_CALLS
 
@@ -63,21 +64,6 @@ typedef struct
 
 static GObjectClass *parent_class = NULL;
 
-enum {
-	STRUT_LEFT = 0,
-	STRUT_RIGHT = 1,
-	STRUT_TOP = 2,
-	STRUT_BOTTOM = 3,
-	STRUT_LEFT_START = 4,
-	STRUT_LEFT_END = 5,
-	STRUT_RIGHT_START = 6,
-	STRUT_RIGHT_END = 7,
-	STRUT_TOP_START = 8,
-	STRUT_TOP_END = 9,
-	STRUT_BOTTOM_START = 10,
-	STRUT_BOTTOM_END = 11
-};
-
 enum {
 	MAGNIFIER_SOURCE_DISPLAY_PROP,
 	MAGNIFIER_TARGET_DISPLAY_PROP,
@@ -130,16 +116,11 @@ gchar* mag_prop_names[MAGNIFIER_CROSSWIRE_COLOR_PROP + 1] = {
 };
 #endif
 
-static int _x_error = 0;
-static int fixes_event_base = 0, fixes_error_base;
-static Display *cursor_client_connection;
-static guint    cursor_client_gsource = 0;
 static Magnifier *_this_magnifier = NULL;
+extern gint       fixes_event_base;
 
-static void magnifier_transform_cursor (Magnifier *magnifier);
 static void magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set);
 static void magnifier_init_window (Magnifier *magnifier, GdkScreen *screen);
-static gboolean magnifier_check_set_struts (Magnifier *magnifier);
 static gboolean magnifier_reset_struts_at_idle (gpointer data);
 static void magnifier_init_window (Magnifier *magnifier, GdkScreen *screen);
 static void magnifier_adjust_source_size (Magnifier *magnifier);
@@ -204,34 +185,11 @@ BONOBO_TYPE_FUNC (MagLoginHelper,
 		  LOGIN_HELPER_TYPE,
 		  mag_login_helper)
 
-gboolean
-magnifier_error_check (void)
-{
-	if (_x_error) {
-		_x_error = 0;
-		return TRUE;
-	}
-	return FALSE;
-}
-
-static int
-magnifier_x_error_handler (Display	 *display,
-			   XErrorEvent *error)
-{
-	if (error->error_code == BadAlloc) {
-		_x_error = error->error_code;
-	}
-	else {
-		return -1;
-	}
-	return 0;
-}
-
 static gboolean
 can_open_display (gchar *display_name)
 {
     Display *d;
-    if (d = XOpenDisplay (display_name))
+    if ((d = XOpenDisplay (display_name)))
     {
 	XCloseDisplay (d);
 	return TRUE;
@@ -290,7 +248,8 @@ magnifier_zoom_regions_mark_dirty (Magnifier *magnifier, GNOME_Magnifier_RectBou
 }
 
 void
-magnifier_set_cursor_from_pixbuf (Magnifier *magnifier, GdkPixbuf *cursor_pixbuf)
+magnifier_set_cursor_from_pixbuf (Magnifier *magnifier,
+				  GdkPixbuf *cursor_pixbuf)
 {
 	GdkPixmap *pixmap, *mask;
 	gint width, height;
@@ -301,89 +260,46 @@ magnifier_set_cursor_from_pixbuf (Magnifier *magnifier, GdkPixbuf *cursor_pixbuf
 		g_object_unref (magnifier->priv->cursor);
 		magnifier->priv->cursor = NULL;
 	}
-	if (drawable && cursor_pixbuf)
-	{
+	if (drawable && cursor_pixbuf) {
 		const gchar *xhot_string = NULL, *yhot_string = NULL;
 		width = gdk_pixbuf_get_width (cursor_pixbuf);
 		height = gdk_pixbuf_get_height (cursor_pixbuf);
 		pixmap = gdk_pixmap_new (drawable, width, height, -1);
 		gc = gdk_gc_new (pixmap);
 		if (GDK_IS_DRAWABLE (pixmap))
-		    gdk_draw_pixbuf (pixmap, gc, cursor_pixbuf, 0, 0, 0, 0, 
-				     width, height,
-				     GDK_RGB_DITHER_NONE, 0, 0);
+			gdk_draw_pixbuf (pixmap, gc, cursor_pixbuf, 0, 0, 0, 0,
+					 width, height, GDK_RGB_DITHER_NONE,
+					 0, 0);
 		else
-		    DBG (g_warning ("empty cursor pixmap created."));
+			DBG (g_warning ("empty cursor pixmap created."));
 		mask = gdk_pixmap_new (drawable, width, height, 1);
-		gdk_pixbuf_render_threshold_alpha (cursor_pixbuf, mask, 0, 0, 0, 0, 
+		gdk_pixbuf_render_threshold_alpha (cursor_pixbuf, mask,
+						   0, 0, 0, 0, 
 						   width, height,
 						   200);
-		gdk_gc_unref (gc);
+		g_object_unref (gc);
 		magnifier->priv->cursor = pixmap;
 		magnifier->priv->cursor_mask = mask;
 		xhot_string = gdk_pixbuf_get_option (cursor_pixbuf,"x_hot");
 		yhot_string = gdk_pixbuf_get_option (cursor_pixbuf,"y_hot");
-		if (xhot_string) magnifier->cursor_hotspot.x = atoi (xhot_string);
-		if (yhot_string) magnifier->cursor_hotspot.y = atoi (yhot_string);
-
+		if (xhot_string)
+			magnifier->cursor_hotspot.x = atoi (xhot_string);
+		if (yhot_string)
+			magnifier->cursor_hotspot.y = atoi (yhot_string);
 		if (pixmap) {
-			gdk_drawable_get_size (pixmap,
-					       &magnifier->priv->cursor_default_size_x,
-					       &magnifier->priv->cursor_default_size_y);
-			magnifier->priv->cursor_hotspot_x = magnifier->cursor_hotspot.x;
-			magnifier->priv->cursor_hotspot_y = magnifier->cursor_hotspot.y;
+			gdk_drawable_get_size (
+				pixmap,
+				&magnifier->priv->cursor_default_size_x,
+				&magnifier->priv->cursor_default_size_y);
+			magnifier->priv->cursor_hotspot_x =
+				magnifier->cursor_hotspot.x;
+			magnifier->priv->cursor_hotspot_y =
+				magnifier->cursor_hotspot.y;
 		}
 	}
 }
 
 
-void
-magnifier_free_cursor_pixels (guchar *pixels, gpointer data)
-{
-    /* XFree (data); FIXME why doesn't this work properly? */
-}
-
-#ifdef HAVE_XFIXES
-static void
-magnifier_cursor_convert_to_rgba (Magnifier *magnifier, XFixesCursorImage *cursor_image)
-{
-	int i, count = cursor_image->width * cursor_image->height;
-	for (i = 0; i < count; ++i) 
-	{
-		guint32 pixval = GUINT_TO_LE (cursor_image->pixels[i]);
-		cursor_image->pixels[i] = pixval;
-	}
-}
-#endif
-
-GdkPixbuf *
-magnifier_get_source_pixbuf (Magnifier *magnifier)
-{
-#ifdef HAVE_XFIXES
-	XFixesCursorImage *cursor_image = XFixesGetCursorImage (cursor_client_connection);
-        GdkPixbuf *cursor_pixbuf = NULL;
-	gchar s[6];
-	if (cursor_image)
-	{
-	        magnifier_cursor_convert_to_rgba (magnifier, cursor_image);
-		cursor_pixbuf = gdk_pixbuf_new_from_data ((guchar *) cursor_image->pixels, 
-							  GDK_COLORSPACE_RGB, 
-							  TRUE, 8,  
-							  cursor_image->width, cursor_image->height,
-							  cursor_image->width * 4, 
-							  magnifier_free_cursor_pixels, 
-							  cursor_image );
-		gdk_pixbuf_set_option (cursor_pixbuf, "x_hot", 
-				       g_ascii_dtostr (s, 6, (gdouble) cursor_image->xhot));
-		gdk_pixbuf_set_option (cursor_pixbuf, "y_hot", 
-				       g_ascii_dtostr (s, 6, (gdouble) cursor_image->yhot));
-	}
-	return cursor_pixbuf;
-#else
-	return NULL;
-#endif
-}
-
 GdkPixbuf *
 magnifier_get_pixbuf_for_name (Magnifier *magnifier, const gchar *cursor_name)
 {
@@ -396,18 +312,16 @@ magnifier_get_pixbuf_for_name (Magnifier *magnifier, const gchar *cursor_name)
 }
 
 void
-magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier, const gchar *cursor_name, 
+magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier,
+				     const gchar *cursor_name, 
 				     gboolean source_fallback)
 {
 	GdkPixbuf *pixbuf;
 	/* search local table; if not found, use source screen's cursor if source_fallback is TRUE */
 	if ((pixbuf = magnifier_get_pixbuf_for_name (magnifier, cursor_name)) == NULL) {
-#ifndef HAVE_XFIXES
-		source_fallback = FALSE;
-#endif
 		if (source_fallback == TRUE)
 		{
-			pixbuf = magnifier_get_source_pixbuf (magnifier);
+			pixbuf = gmag_cursor_get_source_pixbuf (magnifier);
 		}
 		else
 		{
@@ -415,101 +329,11 @@ magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier, const gchar *cursor_n
 		}
 	}
 	magnifier_set_cursor_from_pixbuf (magnifier, pixbuf);
-	if (pixbuf) gdk_pixbuf_unref (pixbuf);
-}
-
-gboolean
-magnifier_cursor_notify (GIOChannel *source, GIOCondition condition, gpointer data)
-{
-#ifdef HAVE_XFIXES
-  XEvent ev;
-  Magnifier *magnifier = (Magnifier *) data;
-  XFixesCursorNotifyEvent  *cev;
-
-  /* TODO: don't ask for name unless xfixes version >= 2 available */
-  do
-  {
-      XNextEvent(cursor_client_connection, &ev);
-      if (ev.type == fixes_event_base + XFixesCursorNotify) 
-      {
-	  cev = (XFixesCursorNotifyEvent *) &ev;
-      }
-  } while (XPending (cursor_client_connection));
-
-  if (magnifier->priv->use_source_cursor) 
-  {
-      GdkPixbuf *cursor_pixbuf = magnifier_get_source_pixbuf (magnifier);
-      magnifier_set_cursor_from_pixbuf (magnifier, cursor_pixbuf);
-      if (cursor_pixbuf) gdk_pixbuf_unref (cursor_pixbuf);
-  }
-  else
-  {
-      magnifier_set_cursor_pixmap_by_name (magnifier, 
-					   gdk_x11_get_xatom_name (cev->cursor_name),
-					   TRUE);
-  }
-
-  magnifier_transform_cursor (magnifier);
-#ifdef DEBUG_CURSOR
-  g_message ("cursor changed: subtype=%d, cursor_serial=%lu, name=[%x] %s\n",
-	     (int) cev->subtype, cev->cursor_serial, (int) cev->cursor_name,
-	     gdk_x11_get_xatom_name (cev->cursor_name));
-#endif
-
-  return TRUE;
-#else
-  return FALSE;
-#endif
-}
-
-gboolean
-magnifier_cursor_notification_init (Magnifier *magnifier)
-{
-#ifdef HAVE_XFIXES
-    GIOChannel *ioc;
-    int fd;
-    Display *dpy;
-    Window rootwin;
-
-    if (!magnifier->source_display) return FALSE;
-
-    if (cursor_client_connection)
-    {
-        /* remove the old watch */
-	if (cursor_client_gsource) 
-	  g_source_remove (cursor_client_gsource);
-	XCloseDisplay (cursor_client_connection);
-    }
-
-    cursor_client_connection = XOpenDisplay (magnifier->source_display_name);
-    rootwin = GDK_WINDOW_XWINDOW (magnifier->priv->root);
-
-    if (!XFixesQueryExtension (cursor_client_connection, &fixes_event_base, &fixes_error_base))
-    {
-	g_warning ("XFixes extension not currently active.\n");
-	return FALSE;
-    }
-    else
-    {
-	XFixesSelectCursorInput (cursor_client_connection, rootwin, XFixesDisplayCursorNotifyMask);
-	fd = ConnectionNumber (cursor_client_connection);
-	ioc = g_io_channel_unix_new (fd);
-	cursor_client_gsource = 
-	    g_io_add_watch (ioc, G_IO_IN | G_IO_HUP | G_IO_PRI | G_IO_ERR, magnifier_cursor_notify, 
-			    magnifier);
-	g_io_channel_unref (ioc); 
-	g_message ("added event source to xfixes cursor-notify connection");
- 	XFlush (cursor_client_connection); 
-   }
-    return TRUE;
-#else
-    g_warning ("this copy of gnome-mag was built without xfixes extension support.\n");
-    return FALSE;
-#endif
+	if (pixbuf) g_object_unref (pixbuf);
 }
 
 void
-magnifier_notify_damage (Magnifier *magnifier, XRectangle *rect)
+magnifier_notify_damage (Magnifier *magnifier, GdkRectangle *rect)
 {
 	GNOME_Magnifier_RectBounds rect_bounds;
 	rect_bounds.x1 = rect->x;
@@ -519,96 +343,29 @@ magnifier_notify_damage (Magnifier *magnifier, XRectangle *rect)
 #undef DEBUG_DAMAGE
 #ifdef DEBUG_DAMAGE
 	g_message ("damage");
-	g_message ("dirty %d, %d to %d, %d", rect_bounds.x1, rect_bounds.y1, rect_bounds.x2, rect_bounds.y2);
+	g_message ("dirty %d, %d to %d, %d", rect_bounds.x1, rect_bounds.y1,
+		   rect_bounds.x2, rect_bounds.y2);
 #endif
 	magnifier_zoom_regions_mark_dirty (magnifier, rect_bounds);
 }
 
-GdkFilterReturn
-magnifier_expose_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
-{
-	Magnifier *magnifier = data;
-	GList *list;
-
-	if (event->any.type == GDK_EXPOSE) 
-	{
-		GNOME_Magnifier_RectBounds rect_bounds;
-		GdkRectangle rect;
-		gdk_region_get_clipbox (event->expose.region, &rect);
-		rect_bounds.x1 = rect.x;
-		rect_bounds.y1 = rect.y;
-		rect_bounds.x2 = rect.x + rect.width;
-		rect_bounds.y2 = rect.y + rect.height;
-		magnifier_zoom_regions_mark_dirty (magnifier, rect_bounds);
-
-		return GDK_FILTER_TRANSLATE;
-	}
-	return GDK_FILTER_CONTINUE;
-}
 static void
 magnifier_set_extension_listeners (Magnifier *magnifier, GdkWindow *root)
 {
-	if (!magnifier_damage_client_init (magnifier))
-	    g_warning ("Damage client hooks were not initialized.");
-	if (!magnifier_cursor_notification_init (magnifier))
-	    g_warning ("Cursor change notification hooks were not initialized.");
+	gmag_gs_client_init (magnifier);
 	magnifier->source_initialized = TRUE;
 }
 
 static void
 magnifier_size_allocate (GtkWidget *widget)
 {
-	magnifier_check_set_struts (_this_magnifier);
+	gmag_gs_check_set_struts (_this_magnifier);
 }
 
 static void
 magnifier_realize (GtkWidget *widget)
 {
-	XWMHints wm_hints;
-	Atom wm_window_protocols[2];
-	Atom wm_type_atoms[1];
-	Atom net_wm_window_type;
-	GdkDisplay *target_display = gdk_drawable_get_display (widget->window);
-	
-	static gboolean initialized = FALSE;
-
-#ifndef MAG_WINDOW_OVERRIDE_REDIRECT	
-	if (!initialized) {
-		wm_window_protocols[0] = gdk_x11_get_xatom_by_name_for_display (target_display,
-										"WM_DELETE_WINDOW");
-		wm_window_protocols[1] = gdk_x11_get_xatom_by_name_for_display (target_display,
-										"_NET_WM_PING");
-		/* use DOCK until Metacity RFE for new window type goes in */
-                wm_type_atoms[0] = gdk_x11_get_xatom_by_name_for_display (target_display,
-									  "_NET_WM_WINDOW_TYPE_DOCK");
-	}
-  
-	wm_hints.flags = InputHint;
-	wm_hints.input = False;
-	
-	XSetWMHints (GDK_WINDOW_XDISPLAY (widget->window),
-		     GDK_WINDOW_XWINDOW (widget->window), &wm_hints);
-	
-	XSetWMProtocols (GDK_WINDOW_XDISPLAY (widget->window),
-			 GDK_WINDOW_XWINDOW (widget->window), wm_window_protocols, 2);
-
-	net_wm_window_type = gdk_x11_get_xatom_by_name_for_display 
-		(target_display, "_NET_WM_WINDOW_TYPE");
-
-	if (net_wm_window_type && wm_type_atoms[0])
-		XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window),
-				 GDK_WINDOW_XWINDOW (widget->window),
-				 net_wm_window_type,
-				 XA_ATOM, 32, PropModeReplace,
-				 (guchar *)wm_type_atoms,
-				 1);
-#else
-#endif
-	/* TODO: make sure this works/is reset if the DISPLAY 
-	 * (as well as the SCREEN) changes.
-	 */
-
-	XSetErrorHandler (magnifier_x_error_handler);
+	gmag_gs_magnifier_realize (widget);
 }
 
 GdkWindow*
@@ -672,39 +429,48 @@ magnifier_get_display_rect_bounds (Magnifier *magnifier, GNOME_Magnifier_RectBou
     }
 }
 
-static void magnifier_adjust_source_size (Magnifier *magnifier)
+static void
+magnifier_adjust_source_size (Magnifier *magnifier)
 {
 	GNOME_Magnifier_RectBounds rect_bounds;	
+	gdouble vfract_top, vfract_bottom, hfract_left, hfract_right;
 	magnifier_get_display_rect_bounds (magnifier, &rect_bounds, FALSE);
-	gdouble vfract, hfract;
-	hfract = (double) (magnifier->target_bounds.x2 - magnifier->target_bounds.x1) / (double) (rect_bounds.x2 - rect_bounds.x1);
-	vfract = (double) (magnifier->target_bounds.y2 - magnifier->target_bounds.y1) / (double) (rect_bounds.y2 - rect_bounds.y1);
-	if (vfract > hfract) /* vertical splitpane, approximately */
+	hfract_left = (double) (magnifier->target_bounds.x1) / (double) rect_bounds.x2;
+	vfract_top = (double) (magnifier->target_bounds.y1) / (double) rect_bounds.y2;
+	hfract_right = (double) (rect_bounds.x2 - magnifier->target_bounds.x2) / (double) rect_bounds.x2;
+	vfract_bottom = (double) (rect_bounds.y2 - magnifier->target_bounds.y2) / (double) rect_bounds.y2;
+	/* we make our 'source' rectangle the largest available subsection which we aren't occupying */
+#ifdef HAVE_COMPOSITE
+	if (!g_getenv ("MAGNIFIER_IGNORE_COMPOSITE")) {
+		magnifier->source_bounds = rect_bounds;
+	} else
+#endif
+	if (MAX (hfract_left, hfract_right) > MAX (vfract_top, vfract_bottom))  /* vertical split, approximately */
 	{
-		if (magnifier->target_bounds.x1 < magnifier->source_bounds.x1) {
+		if (hfract_right > hfract_left) {
 			magnifier->source_bounds.x1 = magnifier->target_bounds.x2;
 			magnifier->source_bounds.x2 = rect_bounds.x2;
 		}
 		else 
 		{
-			magnifier->source_bounds.x1 = 0;
+			magnifier->source_bounds.x1 = rect_bounds.x1;
 			magnifier->source_bounds.x2 = magnifier->target_bounds.x1;
 		}
-		magnifier->source_bounds.y1 = 0;
+		magnifier->source_bounds.y1 = rect_bounds.y1;
 		magnifier->source_bounds.y2 = rect_bounds.y2;
 	}
-	else
+	else /* more-or-less horizontally split */
 	{
-		if (magnifier->target_bounds.y1 < magnifier->source_bounds.y1) {
+		if (vfract_bottom > vfract_top) {
 			magnifier->source_bounds.y1 = magnifier->target_bounds.y2;
 			magnifier->source_bounds.y2 = rect_bounds.y2;
 		}
 		else 
 		{
-			magnifier->source_bounds.y1 = 0;
+			magnifier->source_bounds.y1 = rect_bounds.y1;
 			magnifier->source_bounds.y2 = magnifier->target_bounds.y1;
 		}
-		magnifier->source_bounds.x1 = 0;
+		magnifier->source_bounds.x1 = rect_bounds.x1;
 		magnifier->source_bounds.x2 = rect_bounds.x2;
 	}
 	g_message ("set source bounds to %d,%d; %d,%d", 
@@ -905,7 +671,7 @@ magnifier_recolor_pixbuf (Magnifier *magnifier, GdkPixbuf *pixbuf)
 	}
 }
 
-static void
+void
 magnifier_transform_cursor (Magnifier *magnifier)
 {
 	if (magnifier->priv->cursor) /* don't do this if cursor isn't intialized yet */
@@ -914,6 +680,7 @@ magnifier_transform_cursor (Magnifier *magnifier)
 		int size_x, size_y;
 		GdkPixbuf *scaled_cursor_pixbuf;
 		GdkPixbuf *scaled_mask_pixbuf;
+		GdkPixbuf *scaled_mask_pixbuf_alpha;
 		GdkPixbuf *cursor_pixbuf;
 		GdkPixbuf *mask_pixbuf;
 		GdkPixmap *cursor_pixmap = magnifier->priv->cursor;
@@ -985,12 +752,13 @@ magnifier_transform_cursor (Magnifier *magnifier)
 		}
 		else
 		    DBG (g_warning ("cursor pixmap is non-drawable."));
-		scaled_mask_pixbuf = gdk_pixbuf_add_alpha (
+		scaled_mask_pixbuf_alpha = gdk_pixbuf_add_alpha (
 			scaled_mask_pixbuf, True, 0, 0, 0);
-		gdk_pixbuf_render_threshold_alpha (scaled_mask_pixbuf,
+		gdk_pixbuf_render_threshold_alpha (scaled_mask_pixbuf_alpha,
 						   magnifier->priv->cursor_mask,
 						   0, 0, 0, 0, size_x, size_y,
 						   0x80);
+		g_object_unref (scaled_mask_pixbuf_alpha);
 		g_object_unref (scaled_cursor_pixbuf);
 		g_object_unref (scaled_mask_pixbuf);
 		g_object_unref (mgc);
@@ -1001,19 +769,14 @@ magnifier_transform_cursor (Magnifier *magnifier)
 static void
 magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set)
 {
-	GdkPixbuf	*cursor_pixbuf = NULL;
 	/*
 	 * we check the cursor-set property string here,
 	 * and create/apply the appropriate cursor settings
 	 */
 	magnifier->cursor_set = cursor_set;
-#ifdef HAVE_XFIXES	
 	magnifier->priv->use_source_cursor = 
 	    (!strcmp (cursor_set, "default") && 
 	     (fixes_event_base != 0));
-#else
-	magnifier->priv->use_source_cursor = FALSE;
-#endif
 	if (magnifier->priv->use_source_cursor) return;
 
 	if (!strcmp (magnifier->cursor_set, "none")) {
@@ -1071,101 +834,15 @@ magnifier_reset_struts_at_idle (gpointer data)
 	if (data)
 	{
 		Magnifier *magnifier = MAGNIFIER (data);
-		if (magnifier->priv && GTK_WIDGET_REALIZED (magnifier->priv->w) && 
-		    magnifier_check_set_struts (magnifier))
-		{
+		if (magnifier->priv &&
+		    GTK_WIDGET_REALIZED (magnifier->priv->w) && 
+		    gmag_gs_check_set_struts (magnifier)) {
 			return FALSE;
 		}
 	}
 	return TRUE;
 }
 
-static gboolean
-magnifier_check_set_struts (Magnifier *magnifier)
-{
-	/* TODO: don't do this if we're using Composite */
-	if (magnifier &&
-	    magnifier->priv && magnifier->priv->w && GTK_WIDGET_REALIZED (magnifier->priv->w) &&
-	    magnifier->priv->w->window) 
-	{
-		Atom atom_strut = gdk_x11_get_xatom_by_name ("_NET_WM_STRUT");
-		Atom atom_strut_partial = gdk_x11_get_xatom_by_name ("_NET_WM_STRUT_PARTIAL");
-		guint32 struts[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-		GtkWidget *widget = magnifier->priv->w;
-		gint width = gdk_screen_get_width (
-			gdk_display_get_screen (magnifier->target_display,
-						magnifier->target_screen_num));
-		gint height = gdk_screen_get_height (
-			gdk_display_get_screen (magnifier->target_display,
-						magnifier->target_screen_num));
-
-		gint right_margin, left_margin, top_margin, bottom_margin;
-		gint wx, wy, ww, wh;
-
-		gtk_window_get_position (GTK_WINDOW (magnifier->priv->w), &wx, &wy);
-		gtk_window_get_size (GTK_WINDOW (magnifier->priv->w), &ww, &wh);
-
-		left_margin = wx;
-		right_margin = (width - ww) - wx;
-		top_margin = wy;
-		bottom_margin = (height - wh) - wy;
-
-		/* set the WM_STRUT properties on the appropriate side */
-		if (bottom_margin > top_margin && 
-		    bottom_margin > left_margin &&
-		    bottom_margin > right_margin)
-		{
-			struts[STRUT_TOP] = wh + wy;
-			struts[STRUT_TOP_START] = wx;
-			struts[STRUT_TOP_END] = wx + ww;
-		} 
-		else if (top_margin > bottom_margin && 
-			 top_margin > left_margin &&
-			 top_margin > right_margin)
-		{
-			struts[STRUT_BOTTOM] = height - wy;
-			struts[STRUT_BOTTOM_START] = wx;
-			struts[STRUT_BOTTOM_END] = wx + ww;
-		}
-		else if (right_margin > left_margin &&
-			 right_margin > top_margin &&
-			 right_margin > bottom_margin)
-		{
-			struts[STRUT_LEFT] = wx;
-			struts[STRUT_LEFT_START] = wy;
-			struts[STRUT_LEFT_END] = wh + wy;
-		}
-		else 
-		{
-			struts[STRUT_RIGHT] = width - wx;
-			struts[STRUT_RIGHT_START] = wy;
-			struts[STRUT_RIGHT_END] = wy + wh;
-		}
-		
-		gdk_error_trap_push ();
-		XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window), 
-				 GDK_WINDOW_XWINDOW (widget->window), 
-				 atom_strut,
-				 XA_CARDINAL, 32, PropModeReplace,
-				 (guchar *) &struts, 4);
-		XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window), 
-				 GDK_WINDOW_XWINDOW (widget->window), 
-				 atom_strut_partial,
-				 XA_CARDINAL, 32, PropModeReplace,
-				 (guchar *) &struts, 12); 
-		gdk_error_trap_pop ();
-
-#ifdef DEBUG_STRUTS
-		g_message ("struts TOP %d (%d - %d)", struts[STRUT_TOP], struts[STRUT_TOP_START], struts[STRUT_TOP_END]);
-		g_message ("struts BOTTOM %d (%d - %d)", struts[STRUT_BOTTOM], struts[STRUT_BOTTOM_START], struts[STRUT_BOTTOM_END]);
-		g_message ("struts LEFT %d (%d - %d)", struts[STRUT_LEFT], struts[STRUT_LEFT_START], struts[STRUT_LEFT_END]);
-		g_message ("struts RIGHT %d (%d - %d)", struts[STRUT_RIGHT], struts[STRUT_RIGHT_START], struts[STRUT_RIGHT_END]);
-#endif
-		return TRUE;
-	}
-	return FALSE;
-}
-
 static void
 magnifier_get_property (BonoboPropertyBag *bag,
 			BonoboArg *arg,
@@ -1174,7 +851,6 @@ magnifier_get_property (BonoboPropertyBag *bag,
 			gpointer user_data)
 {
 	Magnifier *magnifier = user_data;
-	GNOME_Magnifier_RectBounds rect_bounds;
 	int csize = 0;
 
 	DBG (fprintf (stderr, "Get property: \t%s\n", mag_prop_names[arg_id]));
@@ -1246,7 +922,6 @@ magnifier_set_property (BonoboPropertyBag *bag,
 			gpointer user_data)
 {
 	Magnifier *magnifier = user_data;
-	GNOME_Magnifier_RectBounds rect_bounds;
 	gchar *full_display_string;
 
 	switch (arg_id) {
@@ -1268,12 +943,34 @@ magnifier_set_property (BonoboPropertyBag *bag,
 			    gdk_display_get_screen (
 				magnifier->source_display,
 				magnifier->source_screen_num));
-	             /* attach listeners for DAMAGE, "dirty region", XFIXES cursor changes */
+
+		    /* remove the source_drawable, since in the new display the
+		     * composite can be unavailable and the source_drawable is
+		     * where the desktop screen is composed. If this is not
+		     * freed it will be used even if the display doesn't use
+		     * composite what will lead to wrong behavior and crashes.
+		     * If composite is used in the new display this variable is
+		     * re-created */
+		    if (magnifier->priv->source_drawable) {
+			    g_object_unref (magnifier->priv->source_drawable);
+			    magnifier->priv->source_drawable = NULL;
+		    }
+
+		    /* If the COMPOSITE extension was not available in the
+		     * previous display the MAGNIFIER_IGNORE_COMPOSITE
+		     * variable is setted. We must unset it to be sure that we
+		     * will make use of this extension if it's present in the
+		     * new display.
+		     */
+		    g_unsetenv ("MAGNIFIER_IGNORE_COMPOSITE");
+
+		    /* attach listeners for DAMAGE, "dirty region", XFIXES
+		     * cursor changes */
 	            magnifier_set_extension_listeners (magnifier, magnifier_get_root (magnifier));
 		    magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
 		    magnifier_zoom_regions_restore (magnifier, zoom_region_params);
 		    magnifier_warp_cursor_to_screen (magnifier);
-		    magnifier_check_set_struts (magnifier);
+		    gmag_gs_check_set_struts (magnifier);
 		}
 		DBG(fprintf (stderr, "Set source display: \t%s\n", full_display_string));
 		break;
@@ -1317,13 +1014,25 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		    }
 		    magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
 		    magnifier_init_cursor_set (magnifier, magnifier->cursor_set); /* needed to reset pixmaps */
-		    gtk_window_move (GTK_WINDOW (magnifier->priv->w),
-				     magnifier->target_bounds.x1,
-				     magnifier->target_bounds.y1);
+		    if (magnifier->priv->overlay)
+			    gdk_window_move (magnifier->priv->overlay,
+					     magnifier->target_bounds.x1,
+					     magnifier->target_bounds.y1);
+		    else
+			    gtk_window_move (GTK_WINDOW (magnifier->priv->w),
+					     magnifier->target_bounds.x1,
+					     magnifier->target_bounds.y1);
 		    
 		    if ((magnifier->target_bounds.x2 - magnifier->target_bounds.x1 > 0) &&
 			(magnifier->target_bounds.y2 - magnifier->target_bounds.y1) > 0)
 		    {
+			    if (magnifier->priv->overlay)
+				    gdk_window_resize (
+					    magnifier->priv->overlay,
+					    magnifier->target_bounds.x2 -
+					    magnifier->target_bounds.x1,
+					    magnifier->target_bounds.y2 -
+					    magnifier->target_bounds.y1);
 			gtk_window_resize (GTK_WINDOW (magnifier->priv->w),
 				       magnifier->target_bounds.x2 - magnifier->target_bounds.x1,
 				       magnifier->target_bounds.y2 - magnifier->target_bounds.y1);
@@ -1333,7 +1042,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		    /* N. B. we don't reset the target bounds to the limits of the new display, because */
 		    /* doing so would override the client-specified magnifier size */
 		    /* magnifier_get_display_rect_bounds (magnifier, &magnifier->target_bounds, TRUE); */
-		    magnifier_check_set_struts (magnifier);
+		    gmag_gs_check_set_struts (magnifier);
 		}
 		DBG(fprintf (stderr, "Set target display: \t%s (screen %d)\n", 
 			      full_display_string, magnifier->target_screen_num));
@@ -1350,15 +1059,25 @@ magnifier_set_property (BonoboPropertyBag *bag,
 	        magnifier->target_bounds = BONOBO_ARG_GET_GENERAL (arg,
 								   TC_GNOME_Magnifier_RectBounds,
 								   GNOME_Magnifier_RectBounds,
-								   NULL);
-		gtk_window_move (GTK_WINDOW (magnifier->priv->w),
-				 magnifier->target_bounds.x1,
-				 magnifier->target_bounds.y1);
+				
+				   NULL);
+		if (magnifier->priv->overlay)
+			gdk_window_move_resize (magnifier->priv->overlay,
+						magnifier->target_bounds.x1,
+						magnifier->target_bounds.y1,
+						magnifier->target_bounds.x2 -
+						magnifier->target_bounds.x1,
+						magnifier->target_bounds.y2 -
+						magnifier->target_bounds.y1);
+		else
+			gtk_window_move (GTK_WINDOW (magnifier->priv->w),
+					 magnifier->target_bounds.x1,
+					 magnifier->target_bounds.y1);
 		
 		gtk_window_resize (GTK_WINDOW (magnifier->priv->w),
 				   magnifier->target_bounds.x2 - magnifier->target_bounds.x1,
 				   magnifier->target_bounds.y2 - magnifier->target_bounds.y1);
-		magnifier_check_set_struts (magnifier);
+		gmag_gs_check_set_struts (magnifier);
 		DBG(fprintf (stderr, "Set target size: \t%d,%d to %d,%d\n", 
 			      magnifier->target_bounds.x1, magnifier->target_bounds.y1, magnifier->target_bounds.x2, magnifier->target_bounds.y2));
 		if (!strcmp (magnifier->target_display_name, magnifier->source_display_name) && 
@@ -1415,6 +1134,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		break;
 	default:
 		bonobo_exception_set (ev, ex_Bonobo_PropertyBag_NotFound);
+		break;
 	};
 }
 
@@ -1447,12 +1167,18 @@ impl_magnifier_set_source_display (PortableServer_Servant servant,
 {
 	Magnifier *magnifier = MAGNIFIER (bonobo_object_from_servant (servant));
 	BonoboArg *arg = bonobo_arg_new (BONOBO_ARG_STRING);
-	BONOBO_ARG_SET_STRING (arg, display);
+	gchar *full_display_string;
+
+	full_display_string = g_strdup (display);
+	if (strcmp (full_display_string, "") == 0)
+		full_display_string = (gchar *) g_getenv ("DISPLAY");
+
+	BONOBO_ARG_SET_STRING (arg, full_display_string);
 	
-	DBG (fprintf (stderr, "Set source display: \t%s\n", display));
-
-	if (strcmp (display, magnifier->source_display_name)) {
+	DBG (fprintf (stderr, "Set source display: \t%s\n",
+		      full_display_string));
 
+	if (strcmp (full_display_string, magnifier->source_display_name)) {
 	    magnifier_set_property (magnifier->property_bag,
 				    arg,
 				    MAGNIFIER_SOURCE_DISPLAY_PROP,
@@ -1462,7 +1188,7 @@ impl_magnifier_set_source_display (PortableServer_Servant servant,
 	else
 	{
 	    DBG (fprintf (stderr, "Attempt to set source to same value as previous: %s\n",
-			  display));
+			  full_display_string));
 	}
 	bonobo_arg_release (arg);
 }
@@ -1474,11 +1200,18 @@ impl_magnifier_set_target_display (PortableServer_Servant servant,
 {
 	Magnifier *magnifier = MAGNIFIER (bonobo_object_from_servant (servant));
 	BonoboArg *arg = bonobo_arg_new (BONOBO_ARG_STRING);
-	BONOBO_ARG_SET_STRING (arg, display);
-	
-	DBG (fprintf (stderr, "Set target display: \t%s\n", display));
+	gchar *full_display_string;
 
-	if (strcmp (display, magnifier->target_display_name)) 
+	full_display_string = g_strdup (display);
+	if (strcmp (full_display_string, "") == 0)
+		full_display_string = (gchar *) g_getenv ("DISPLAY");
+
+	BONOBO_ARG_SET_STRING (arg, full_display_string);
+	
+	DBG (fprintf (stderr, "Set target display: \t%s\n",
+		      full_display_string));
+
+	if (strcmp (full_display_string, magnifier->target_display_name)) 
 	{
 	    magnifier_set_property (magnifier->property_bag,
 				    arg,
@@ -1489,7 +1222,7 @@ impl_magnifier_set_target_display (PortableServer_Servant servant,
 	else
 	{
 	    DBG (fprintf (stderr, "Attempt to set target to same value as previous: %s\n",
-			  display));
+			  full_display_string));
 	}
 	bonobo_arg_release (arg);
 }
@@ -1541,6 +1274,7 @@ impl_magnifier_create_zoom_region (PortableServer_Servant servant,
 	zoom_region->priv->parent = magnifier;
 
 	retval = BONOBO_OBJREF (zoom_region);
+
 	/* XXX: should check ev after each call, below */
 	CORBA_exception_init (ev);
 	GNOME_Magnifier_ZoomRegion_setMagFactor (retval, zx, zy, ev);
@@ -1590,7 +1324,7 @@ impl_magnifier_add_zoom_region (PortableServer_Servant servant,
 
 	/* FIXME: this needs proper lifecycle management */
 	magnifier->zoom_regions = g_list_append (magnifier->zoom_regions, region);
-	magnifier_check_set_struts (magnifier);
+	gmag_gs_check_set_struts (magnifier);
 
 	return CORBA_TRUE;
 }
@@ -1876,14 +1610,25 @@ magnifier_init (Magnifier *magnifier)
 	magnifier->crosswire_clip = FALSE;
 	magnifier->cursor_hotspot.x = 0;
 	magnifier->cursor_hotspot.y = 0;
+	magnifier->target_bounds.x1 = 0;
+	magnifier->target_bounds.y1 = 0;
+	magnifier->target_bounds.x2 = 0;
+	magnifier->target_bounds.y2 = 0;
 	magnifier->priv->cursor = NULL;
 	magnifier->priv->w = NULL;
 	magnifier->priv->use_source_cursor = TRUE;
 	magnifier->priv->cursorlist = NULL;
+	magnifier->priv->source_drawable = NULL;
+	magnifier->priv->overlay = NULL;
 	magnifier_init_window (magnifier, 
 			       gdk_display_get_screen (magnifier->target_display, 
 						       magnifier->target_screen_num));
 	magnifier_init_cursor_set (magnifier, "default");
+
+	mag_timing.process = g_timer_new ();
+	mag_timing.frame = g_timer_new ();
+	mag_timing.scale = g_timer_new ();
+	mag_timing.idle = g_timer_new ();
 #ifdef DEBUG_CLIENT_CALLS
 	client_debug = (g_getenv ("MAG_CLIENT_DEBUG") != NULL);
 #endif
@@ -1907,6 +1652,12 @@ magnifier_get_cursor (Magnifier *magnifier)
 		    magnifier_transform_cursor (magnifier);
 		}
 		g_free (default_cursor_filename);
+	    } else {
+		GdkPixbuf *cursor_pixbuf = gmag_cursor_get_source_pixbuf (
+			magnifier);
+		magnifier_set_cursor_from_pixbuf (magnifier, cursor_pixbuf);
+		if (cursor_pixbuf) g_object_unref (cursor_pixbuf);
+		magnifier_transform_cursor (magnifier);
 	    }
         }
 	return magnifier->priv->cursor;
@@ -1933,10 +1684,17 @@ magnifier_new (gboolean override_redirect)
 	
 	ret = bonobo_activation_active_server_register (
 		MAGNIFIER_OAFIID, BONOBO_OBJREF (mag));
-	if (ret != Bonobo_ACTIVATION_REG_SUCCESS)
-	    g_error ("Error registering magnifier server.\n");
+	if (ret != Bonobo_ACTIVATION_REG_SUCCESS) {
+	    if ( ret == Bonobo_ACTIVATION_REG_ALREADY_ACTIVE)
+	    {
+		printf("An instance of magnifier is already active. Exiting Program.\n");
+		exit(0);
+	    } else
+		g_error ("Error registering magnifier server.\n");
+	}
 
-	g_idle_add (magnifier_reset_struts_at_idle, mag);
+	g_timeout_add (500, magnifier_reset_struts_at_idle, mag);
+	g_timeout_add (500, gmag_gs_reset_overlay_at_idle, mag);
 
 	return mag;
 }
