@@ -46,6 +46,7 @@ typedef struct {
 	BonoboPropertyBag *properties;
 	gboolean invert;
 	gboolean is_managed;
+	gboolean draw_cursor;
 	gboolean cache_source;
 	gchar *smoothing;
 	gfloat contrast;
@@ -54,11 +55,15 @@ typedef struct {
 	gfloat contrast_r;
 	gfloat contrast_g;
 	gfloat contrast_b;
+	gfloat bright_r;
+	gfloat bright_g;
+	gfloat bright_b;
 	gint border_size;
 	guint32 border_color; /* A-RGB, 8 bits each, MSB==alpha */
 	gint x_align_policy;  /* TODO: enums here */
 	gint y_align_policy;
 	GNOME_Magnifier_ZoomRegion_ScrollingPolicy smooth_scroll_policy;
+	GNOME_Magnifier_ZoomRegion_ColorBlindFilter color_blind_filter;
         /* bounds of viewport, in target magnifier window coords */
 	GNOME_Magnifier_RectBounds roi;
 	GNOME_Magnifier_RectBounds viewport; 
@@ -68,6 +73,7 @@ typedef struct {
 	gboolean timing_output;
 	gint timing_pan_rate;
 	gboolean exit_magnifier;
+        gboolean poll_mouse;
 #ifdef ZOOM_REGION_DEBUG
         gboolean alive;
 #endif
